package GaiaLBS::DataStore::Redis;

use utf8;
use feature 'say';
use Data::Dumper::Simple;

use common::sense;
use AnyEvent::Redis;
use AnyEvent::CallbackStack;

my $DEBUG	= 1;
my $NOTICE	= 0;
$| = 1;


sub new {
	my ($class, %args) = @_;
	
	my $self  = {};
	
	$self->{redis} = delete $args{redis} // AnyEvent::Redis->new(%args)
		or die "Cannot create AnyEvent::Redis object :$!";
	$self->{redis_args} = \%args;
	
	bless ($self, $class);
	return $self;
}


# FIXME: range check
# FIXME: rewrite by using redis->eval LUA language
# FIXME: croak/return when no element in any of zrangebyscore
# input:
#	apikey => 'xxxxxxx', layer => '123qwe', name => ‘node_name’ 
#	area: (x1, y1, z1, x2, y2, z2)
# ouput:
#	return_code
#	err_msg
sub query_square {
	say 'query_square() begin' if $DEBUG;
	my $self = shift;
	my $cb = pop if "CODE" eq ref $_[-1];
	my $node_info = shift;
	my ($xstart, $ystart, $zstart, $xstop, $ystop, $zstop) = @_;
	
	my $redis = $self->{redis};
	my $return_code;
	
	#
	# prepare the index list
	#
	my $key_prefix = 'index:'.$node_info->{apikey}.':'.$node_info->{layer}.':loc';
	my $node_key = 'node:' . $node_info->{apikey} . ':' . $node_info->{layer} . ':' . $node_info->{name};
	
	my $cv_indexlist = AnyEvent->condvar;
	my %index_list;
	$index_list{'errors'} = [];
	$index_list{'list'} = [];
	
	$cv_indexlist->begin (sub { shift->send (\%index_list) });
	
	# query x in range
	$cv_indexlist->begin;
	$redis->zrangebyscore(
		"$key_prefix:x:all",
		$xstart, $xstop,
		sub {
			my ($result, $err) = @_;
			say "fetched $key_prefix:x:all\n".Dumper(@_) if $DEBUG;
			say Dumper(@$result) if $DEBUG;
			@$result ? push @index_list{'list'}, $node_key.'.nearby.x' : push @index_list{'errors'}, "no result in: $key_prefix:x:all";
			push @index_list{'errors'}, $err if $err;
			say Dumper(%index_list) if $DEBUG;
			$cv_indexlist->end if $err;
			
			if ( @{$index_list{'errors'}} ) {
				$cv_indexlist->end;
			} else {
				$redis->sunionstore(
					$node_key.'.nearby.x',
					@$result,
					sub {
						my ($result, $err) = @_;
						push @index_list{'errors'}, $err if $err;
						say Dumper(%index_list) if $DEBUG;
						$cv_indexlist->end;
					}
				);
			}
		}
	);
	
	# query y in range
	$cv_indexlist->begin;
	$redis->zrangebyscore(
		"$key_prefix:y:all",
		$ystart, $ystop,
		sub {
			my ($result, $err) = @_;
			say "fetched $key_prefix:y:all\n".Dumper(@_) if $DEBUG;
			say Dumper(@$result) if $DEBUG;
			@$result ? push @index_list{'list'}, $node_key.'.nearby.y' : push @index_list{'errors'}, "no result in: $key_prefix:y:all";
			push @index_list{'errors'}, $err if $err;
			say Dumper(%index_list) if $DEBUG;
			$cv_indexlist->end if $err;
			
			if ( @{$index_list{'errors'}} ) {
				$cv_indexlist->end;
			} else {
				$redis->sunionstore(
					$node_key.'.nearby.y',
					@$result,
					sub {
						my ($result, $err) = @_;
						push @index_list{'errors'}, $err if $err;
						say Dumper(%index_list) if $DEBUG;
						$cv_indexlist->end;
					}
				);
			}
		}
	);
	
	# Do Not query if (z1, z2) are both 0
	$cv_indexlist->begin if ( $zstart or $zstop );
	$redis->zrangebyscore(
		"$key_prefix:z:all",
		$zstart, $zstop,
		sub {
			my ($result, $err) = @_;
			say "fetched $key_prefix:z:all\n".Dumper(@_) if $DEBUG;
			say Dumper(@$result) if $DEBUG;
			@$result ? push @index_list{'list'}, $node_key.'.nearby.z' : push @index_list{'errors'}, "no result in: $key_prefix:z:all";
			push @index_list{'errors'}, $err if $err;
			say Dumper(%index_list) if $DEBUG;
			$cv_indexlist->end if $err;
			
			if ( @{$index_list{'errors'}} ) {
				$cv_indexlist->end;
			} else {
				$redis->sunionstore(
					$node_key.'.nearby.z',
					@$result,
					sub {
						my ($result, $err) = @_;
						push @index_list{'errors'}, $err if $err;
						say Dumper(%index_list) if $DEBUG;
						$cv_indexlist->end;
					}
				);
			}
		}
	) if ( $zstart or $zstop );
	
	$cv_indexlist->end;
	
	
	#
	# get the node list
	#
	my $cv_nodelist = AE::cv;
	my %node_list;
	$node_list{'errors'} = [];
	$node_list{'list'} = [];
	
	$cv_indexlist->cb( sub {
		say "get the node list" if $DEBUG;
		my $index_list = $_[0]->recv;
		say Dumper($index_list) if $DEBUG;
		
		#~ $cv_nodelist->send( \@{[ undef, $index_list->{'errors'} ]} ) if @$index_list->{'errors'};
		
		say 'List: '. Dumper( @{ $index_list->{'list'} } ) if $DEBUG;
		
		if (@{ $index_list->{'list'} }) {
			$redis->sinterstore( $node_key .'.nearby', @{ $index_list->{'list'} }, sub {
				my ($result, $err) = @_;
				$cv_nodelist->send( undef, "no result in area" ) if $err or not $result;
				say Dumper $result if $DEBUG;
				
				$redis->smembers( $node_key .".nearby", sub { $cv_nodelist->send( @_ ) } );
				say Dumper $cv_nodelist if $DEBUG;
			});
		} else {
			$cv_nodelist->send( undef, "no result in area" );
		}
	});
	
	
	#
	# get nodes info
	#
	my $cv_nodesinfo = AE::cv;
	my @all_nodes;
	my @error;
	
	$cv_nodelist->cb( sub {
		say "get nodes info" if $DEBUG;
		my($result, $err) = $_[0]->recv;
		push @error, $err if $err;
		say Dumper $result if $DEBUG;
		
		#~ return $cv_nodesinfo->send( undef, $err ) if $err or not $result;
		
		$cv_nodesinfo->begin (sub { shift->send (\@all_nodes, \@error) });
		for my $node_key (@$result) {
			say $node_key if $DEBUG;
			
			$cv_nodesinfo->begin;
			
			my %tmp_info;
			(undef, $tmp_info{'apikey'}, $tmp_info{'layer'}, $tmp_info{'name'} ) = split/:/,$node_key;
			
			$self->query_node(
				\%tmp_info,
				sub {
					my($result, $err) = @_;
					push @error, @$err if @$err;
					push @all_nodes, $result;
					$cv_nodesinfo->end;
				}
			);
			
			
			
		}
		$cv_nodesinfo->end;
	});
	
	
	#
	# delete temp data
	#
	my $cv_final = AE::cv;
	
	$cv_nodesinfo->cb( sub {
		say "clear temp data" if $DEBUG;
		#~ my($result, $err) = $_[0]->recv;
		#~ push @error, $err if $err;
		#~ say Dumper $result if $DEBUG;
		
		#~ return $cv_nodesinfo->send( undef, $err ) if $err or not $result;
		
		$cv_final->begin (sub { shift->send (\@all_nodes, \@error) });
		$redis->del( $node_key .'.nearby', sub {my($result, $err) = @_; push @error, $err if $err; $cv_final->end} );
		$redis->del( $node_key .'.nearby.x', sub {my($result, $err) = @_; push @error, $err if $err; $cv_final->end} );
		$redis->del( $node_key .'.nearby.y', sub {my($result, $err) = @_; push @error, $err if $err; $cv_final->end} );
		$redis->del( $node_key .'.nearby.z', sub {my($result, $err) = @_; push @error, $err if $err; $cv_final->end} );
		$cv_final->end;
	});
	
	defined $cb ? $cv_final->cb( sub { $cb->( $_[0]->recv ) } ) : return $cv_final;
}


# FIXME: delete node from previous list
# input:
#	apikey => 'xxxxxxx', layer => '123qwe', name => ‘node_name’ 
#	loc: (x, y, z)
#	movement: ( (x1, y1, z1, x2, y2, z2), speed)
#	orientation: (360, 360)
# ouput:
#	return_code
#	err_msg
sub store_pos {
	my $self = shift;
	my $cb = pop if "CODE" eq ref $_[-1];
	my $node_info = shift;
	my ($loc, $move, $orien) = @_;
	my $redis = $self->{redis};
	
	my %cv;
	my $node_prev_exist;
	my %node_prev_info;
	
	say Dumper $move if $DEBUG;
	
	my $vector = $move->[0];
	my $speed = $move->[1];
	my $node_key = 'node:' . $node_info->{apikey} . ':' . $node_info->{layer} . ':' . $node_info->{name};
	
	
	say "GaiaLBS::DataStore::Redis NOW!!" if $DEBUG;
	say Dumper ($node_key, $node_info, $loc, $move, $orien) if $DEBUG;
	say Dumper ($vector, $speed) if $DEBUG;
	
	
	my @error;
	$cv{'final'} = AnyEvent->condvar;
	
	
	# FIXME: get previous location, delete self from previous list
	my $cs = AnyEvent::CallbackStack->new;
	#~ $cs->start(%$node_info);
	$cs->start();
	
	# check key existance
	$cs->add( sub {
		say "START check keys exists" if $DEBUG;
		$redis->exists(
			$node_key,
			sub {
				my ($result, $err) = @_;
				push @error, $err if $err;
				$node_prev_exist = $result;
				say "KEY". ($node_prev_exist ? "EXIST: " : " NOT EXIST: ").$node_key if $NOTICE;
				$cs->next;
			},
		);
	});
	
	$cs->add( sub {
		if ($node_prev_exist) {
			say "Getting previous node info" if $DEBUG;
			# get original data
			
			say "START query_node in store_pos" if $DEBUG;
			$self->query_node(
				$node_info,
				sub { %node_prev_info = %{$_[0]}; say Dumper %node_prev_info if $DEBUG; $cs->next },
			);
			
		} else {
			say "New node" if $DEBUG;
			$cs->next
		};
	});
	
	#~ $cs->add( sub {
		#~ $self->_cleanup_info(
			#~ %$node_info,
			#~ sub {
				#~ my ($result, $err) = @_;
				#~ 
				#~ say 'Cleanup Result: '.Dumper(@_) if $DEBUG;
				#~ 
				#~ push @error, $err if @$err;
				#~ push @error, 'node revoke failed' if @$err;
				#~ $cs->next;
			#~ }
		#~ );
	#~ });
	
	
	$cs->add( sub {
		$cv{'node_basic'} = AnyEvent->condvar;
		$cv{'node_basic'}->begin (sub { shift->send ( '', \@error ) });
		
		($loc->[0], $loc->[1], $loc->[2]) = ($loc->[0] // 0, $loc->[1] // 0, $loc->[2] // 0);
		
		# store node data
		$cv{'node_basic'}->begin;
		$redis->hmset(
			$node_key,
			ident	=> $node_info->{apikey}.':'.$node_info->{layer}.':'.$node_info->{name},
			apikey	=> $node_info->{apikey},
			layer	=> $node_info->{layer},
			name	=> $node_info->{name},
			x	=> $loc->[0],
			y	=> $loc->[1],
			z	=> $loc->[2],
			sub {
				my ($result, $err) = @_;
				push @error, $err if $err;
				warn "SET $node_key / $err" if $NOTICE;
				$cv{'node_basic'}->end;
			},
		);
		
		# optional values
		my %node_v;
		if ( grep {defined $_} @$orien ) {
			$node_v{'orien_h'}	= $orien->[0] // 0;
			$node_v{'orien_v'}	= $orien->[1] // 0;
		}
		
		if ( grep {defined $_} @$vector ) {
			$node_v{'sx'}		= $vector->[0] // 0;
			$node_v{'sy'}		= $vector->[1] // 0;
			$node_v{'sz'}		= $vector->[2] // 0;
			$node_v{'dx'}		= $vector->[3] // 0;
			$node_v{'dy'}		= $vector->[4] // 0;
			$node_v{'dz'}		= $vector->[5] // 0;
		}
		
		if ( defined $speed ) {
			$node_v{'speed'}	= $speed // 0;
		}
		
		say Dumper(%node_v) if $DEBUG;
		
		$cv{'node_basic'}->begin if keys %node_v;
		$redis->hmset(
			"$node_key:v",
			%node_v,
			sub {
				my ($result, $err) = @_;
				push @error, $err if $err;
				warn "SET $node_key:v / $err" if $NOTICE;
				$cv{'node_basic'}->end;
			},
		) if keys %node_v;
		$cv{'node_basic'}->end;
		
		$cv{'node_basic'}->cb(sub{$cs->next});
	});
	
	
	# add to apikey:layer.node.list
	# already done in register
	
	#~ my $list_prefix = $node_info->{apikey}.':'.$node_info->{layer};
	#~ $cs->add( sub {
		#~ return if $cv{'final'}->ready;
		#~ 
		#~ say 'adding to apikey:layer.node.list' if $DEBUG;
		#~ 
		#~ my ($result, $err) = shift->recv;
		#~ $cv{'final'}->send( $result, $err ) if @$err;
		#~ return $cs->next( $result, $err ) if @$err;
		#~ 
		#~ $redis->sadd( "$list_prefix.node.list", $node_info->{name}, sub {
			#~ my @error = @$err;
			#~ my ($result, $err) = @_;
			#~ push @error, $err if $err;
			#~ 
			#~ say 'added to apikey:layer.node.list' if $DEBUG;
			#~ 
			#~ $cs->next( '', \@error );
		#~ } );
	#~ } );
	
	
	# add to list:apikey:layer:loc:x.123 / list:apikey:layer:loc:y.123 / list:apikey:layer:loc:z.123
	my $list_prefix = 'list:'.$node_info->{apikey}.':'.$node_info->{layer}.':loc';
	$cs->add( sub {
		return if $cv{'final'}->ready;
		
		my ($result, $err) = shift->recv;
		$cv{'final'}->send( $result, $err ) if @$err;
		return $cs->next( $result, $err ) if @$err;
		
		if ( $node_prev_exist ) {
			my @error = @$err;
			
			say "Not first time, move between lists" if $DEBUG;
			
			my $cv_local = AE::cv;
			$cv_local->begin (sub { shift->send ( '', \@error ) });
			$cv_local->begin;
			$redis->smove(
				"$list_prefix:x.".(int $node_prev_info{'loc'}[0]),
				"$list_prefix:x.".(int $loc->[0]),
				$node_key,
				sub { my ($result, $err) = @_; push @error, $err if $err; $cv_local->end; }
			);
			$cv_local->begin;
			$redis->smove(
				"$list_prefix:y.".(int $node_prev_info{'loc'}[1]),
				"$list_prefix:y.".(int $loc->[1]),
				$node_key,
				sub { my ($result, $err) = @_; push @error, $err if $err; $cv_local->end; }
			);
			$cv_local->begin;
			$redis->smove(
				"$list_prefix:z.".(int $node_prev_info{'loc'}[2]),
				"$list_prefix:z.".(int $loc->[2]),
				$node_key,
				sub { my ($result, $err) = @_; push @error, $err if $err; $cv_local->end; }
			);
			$cv_local->end;
			say "SET $list_prefix: x.".(int $loc->[0])."/y.".(int $loc->[1])."/z.".(int $loc->[0]) if $NOTICE;
			
			$cv_local->cb( sub {
				return if $cv{'final'}->ready;
				
				say 'from: list:apikey:layer:loc:x.'.(int $node_prev_info{'loc'}[0]).' / list:apikey:layer:loc:y.'.(int $node_prev_info{'loc'}[1]).' / list:apikey:layer:loc:z.'.(int $node_prev_info{'loc'}[2]) if $DEBUG;
				say 'to:   list:apikey:layer:loc:x.'.(int $loc->[0]).' / list:apikey:layer:loc:y.'.(int $loc->[1]).' / list:apikey:layer:loc:z.'.(int $loc->[2]) if $DEBUG;
				
				my ($result, $err) = shift->recv;
				$cv{'final'}->send( $result, $err ) if @$err;
				$cs->next( $result, $err );
			});
		} else {
			my @error = @$err;
			
			say "First Time, add to lists" if $DEBUG;
			
			my $cv_local = AE::cv;
			$cv_local->begin (sub { shift->send ( '', \@error ) });
			$cv_local->begin; $redis->sadd( "$list_prefix:x.".(int $loc->[0]), $node_key, sub { my ($result, $err) = @_; push @error, $err if $err; $cv_local->end; } );
			$cv_local->begin; $redis->sadd( "$list_prefix:y.".(int $loc->[1]), $node_key, sub { my ($result, $err) = @_; push @error, $err if $err; $cv_local->end; } );
			$cv_local->begin; $redis->sadd( "$list_prefix:z.".(int $loc->[2]), $node_key, sub { my ($result, $err) = @_; push @error, $err if $err; $cv_local->end; } );
			$cv_local->end;
			say "SET $list_prefix: x.".(int $loc->[0])."/y.".(int $loc->[1])."/z.".(int $loc->[0]) if $NOTICE;
			
			$cv_local->cb( sub {
				return if $cv{'final'}->ready;
				
				say 'added to list:apikey:layer:loc:x.'.(int $loc->[0]).' / list:apikey:layer:loc:y.'.(int $loc->[1]).' / list:apikey:layer:loc:z.'.(int $loc->[2]) if $DEBUG;
				
				my ($result, $err) = shift->recv;
				$cv{'final'}->send( $result, $err ) if @$err;
				$cs->next( $result, $err );
			});
		}
	});
	
	
	# add to indexes
	$cv{'add2index'} = AnyEvent->condvar;
	$cs->add( sub {
		return if $cv{'final'}->ready;
		
		my ($result, $err) = shift->recv;
		$cv{'final'}->send( $result, $err ) if @$err;
		return $cs->next( $result, $err ) if @$err;
		
		my @error = @$err;
		my $cv_local = AE::cv;
		$cv_local->begin (sub { shift->send ( '', \@error ) });
		
		
		
		say "Doing index operations" if $DEBUG;
		# add to index:apikey:layer:loc:x:all / index:apikey:layer:loc:y:all / index:apikey:layer:loc:z:all
		my $index_prefix = 'index:'.$node_info->{apikey}.':'.$node_info->{layer}.':loc';
		$cv_local->begin; $redis->zadd( "$index_prefix:x:all", (int $loc->[0]), "$list_prefix:x.".(int $loc->[0]), sub { my ($result, $err) = @_; push @error, $err if $err; $cv_local->end; } );	# supposed to use WATCH and MULTI
		$cv_local->begin; $redis->zadd( "$index_prefix:y:all", (int $loc->[1]), "$list_prefix:y.".(int $loc->[1]), sub { my ($result, $err) = @_; push @error, $err if $err; $cv_local->end; } );	# supposed to use WATCH and MULTI
		$cv_local->begin; $redis->zadd( "$index_prefix:z:all", (int $loc->[2]), "$list_prefix:z.".(int $loc->[2]), sub { my ($result, $err) = @_; push @error, $err if $err; $cv_local->end; } );	# supposed to use WATCH and MULTI
		say "SET $index_prefix:x/y/z:all" if $NOTICE;
		
		
		#~ # add to index:apikey:loc:x:all / index:apikey:loc:y:all / index:apikey:loc:z:all
		#~ $index_prefix = 'index:'.$node_info->{apikey}.':loc';
		#~ $cv_local->begin; $redis->zadd( "$index_prefix:x:all", (int $loc->[0]), "$list_prefix:x.".(int $loc->[0]), sub { my ($result, $err) = @_; push @error, $err if $err; $cv_local->end; } );	# supposed to use WATCH and MULTI
		#~ $cv_local->begin; $redis->zadd( "$index_prefix:y:all", (int $loc->[1]), "$list_prefix:y.".(int $loc->[1]), sub { my ($result, $err) = @_; push @error, $err if $err; $cv_local->end; } );	# supposed to use WATCH and MULTI
		#~ $cv_local->begin; $redis->zadd( "$index_prefix:z:all", (int $loc->[2]), "$list_prefix:z.".(int $loc->[2]), sub { my ($result, $err) = @_; push @error, $err if $err; $cv_local->end; } );	# supposed to use WATCH and MULTI
		#~ say "SET $index_prefix:x/y/z:all" if $NOTICE;
		#~ 
		#~ 
		#~ # add to index:loc:x:all / index:loc:y:all / index:loc:z:all
		#~ $index_prefix = "index:loc";
		#~ $cv_local->begin; $redis->zadd( "$index_prefix:x:all", (int $loc->[0]), "$list_prefix:x.".(int $loc->[0]), sub { my ($result, $err) = @_; push @error, $err if $err; $cv_local->end; } );	# supposed to use WATCH and MULTI
		#~ $cv_local->begin; $redis->zadd( "$index_prefix:y:all", (int $loc->[1]), "$list_prefix:y.".(int $loc->[1]), sub { my ($result, $err) = @_; push @error, $err if $err; $cv_local->end; } );	# supposed to use WATCH and MULTI
		#~ $cv_local->begin; $redis->zadd( "$index_prefix:z:all", (int $loc->[2]), "$list_prefix:z.".(int $loc->[2]), sub { my ($result, $err) = @_; push @error, $err if $err; $cv_local->end; } );	# supposed to use WATCH and MULTI
		#~ say "SET $index_prefix:x/y/z:all" if $NOTICE;
		
		
		$cv_local->end;
		
		$cv_local->cb( sub {
			return if $cv{'final'}->ready;
			
			my ($result, $err) = shift->recv;
			$cv{'final'}->send( $result, $err ) if @$err;
			$cs->next( $result, $err );
		});
	});
	
	
	# send back to 'final'
	$cs->last->cb( sub {
		return if $cv{'final'}->ready;
		
		my ($result, $err) = shift->recv;
		
		$result = 'OK' unless @$err;
		$cv{'final'}->send( $result, $err );
	} );
	
	
	
	defined $cb ? $cv{'final'}->cb( sub { $cb->( $_[0]->recv ) } ) : return $cv{'final'};
}


sub query_node {
	my $self = shift;
	my $cb = pop if "CODE" eq ref $_[-1];
	my $node_info = shift;
	my $redis = $self->{redis};
	my @error;
	my %node_data;
	
	say Dumper($node_info) if $DEBUG;
	
	my $cv_final = AnyEvent->condvar;
	
	
	my $node = $node_info->{apikey} . ':' . $node_info->{layer} . ':' . $node_info->{name};
	say "Query NODE: ".$node if ($DEBUG or $NOTICE);
	
	my $cv_local = AnyEvent->condvar;
	$cv_local->begin (sub { shift->send ( \%node_data, \@error ) });
	
	$cv_local->begin;
	$redis->hgetall( 'node:'.$node, sub {
		my ($result, $err) = @_;
		
		say 'get node basic'.Dumper($result, $err) if $DEBUG;
		
		my %result = @$result;
		delete $result{apikey};
		delete $result{layer};
		delete $result{name};
		say 'modified node basic'.Dumper($result) if $DEBUG;
		
		push @error, $err if $err;
		
		#~ @node_data{ keys %result } = values %result;
		$node_data{'ident'} = $result{'ident'};
		if ( defined $result{'x'} or defined $result{'y'} or defined $result{'z'} ) {
			my @loc;
			$loc[0] = int $result{'x'} // 0;
			$loc[1] = int $result{'y'} // 0;
			$loc[2] = int $result{'z'} // 0;
			
			$node_data{'loc'} = \@loc if int @loc;
		}
		
		$cv_local->end;
	});
	
	$cv_local->begin;
	$redis->hgetall( 'node:'.$node.':v', sub {
		my ($result, $err) = @_;
		
		say 'get node :v'.Dumper($result, $err) if $DEBUG;
		
		my %result = @$result;
		push @error, $err if $err;
		
		#~ @node_data{ keys %result } = values %result;
		if ( defined $result{'orien_h'} or defined $result{'orien_v'} ) {
			my @orien;
			$orien[0] = int delete $result{'orien_h'} // 0;
			$orien[1] = int delete $result{'orien_v'} // 0;
			
			$node_data{'orien'} = \@orien;
		}
		
		if ( keys %result ) {
			my @vector;
			push @vector, int $result{$_} // 0 for qw/sx sy sz dx dy dz/;
			my $speed = delete $result{'speed'} if defined $result{'speed'};
			
			$node_data{'move'} = \@{[ \@vector, $speed ]};
		}
		
		
		
		$cv_local->end;
	});
	
	
	$cv_local->end;
	
	$cv_local->cb( sub {
		my ($result, $err) = shift->recv;
		
		say 'Final Node info'.Dumper($result, $err) if $DEBUG;
		
		#~ $result = 'ERROR' if @$err;
		say 'Query Node ERROR' if @$err;
		$cv_final->send( $result, $err );
	});
	
	defined $cb ? $cv_final->cb( sub { $cb->( $_[0]->recv ) } ) : return $cv_final;
}


sub register {
	my $self = shift;
	my $cb = pop if "CODE" eq ref $_[-1];
	my %args = @_;
	my $redis = $self->{redis};
	my @error;
	
	say Dumper(%args) if $DEBUG;
	#~ say "la" x 50;
	push @error, 'no input data' unless @_;
	push @error, 'missing data' unless $args{"apikey"} and $args{"layer"};
	
	
	my $cv_final = AnyEvent->condvar;
	
	my $cv_local = AnyEvent->condvar;
	$cv_local->begin (sub { shift->send ( '', \@error ) });
	
	if ( $args{'name'} ) {
		$cv_local->begin;
		$redis->sadd(
			$args{'apikey'}.':'.$args{'layer'}.'.node.list',
			$args{'name'},
			sub {
				my ($result, $err) = @_;
				push @error, $err if $err;
				push @error, 'node register failed' if $err;
				$cv_local->end;
			}
		);
	} else {
		$cv_local->begin;
		$redis->sadd(
			$args{'apikey'} . '.layer.list',
			$args{'layer'},
			sub {
				my ($result, $err) = @_;
				push @error, $err if $err;
				push @error, 'layer register failed' if $err;
				$cv_local->end;
			}
		);
	}
	
	$cv_local->end;
	
	
	$cv_local->cb( sub {
		return if $cv_final->ready;
		
		my ($result, $err) = shift->recv;
		
		$result = 'OK' unless @$err;
		
		$cv_final->send( $result, $err );
	});
	
	defined $cb ? $cv_final->cb( sub { $cb->( $_[0]->recv ) } ) : return $cv_final;
}


sub valid {
	my $self = shift;
	my $cb = pop if "CODE" eq ref $_[-1];
	my %args = @_;
	my $redis = $self->{redis};
	my @error;
	
	say Dumper(%args) if $DEBUG;
	#~ say "la" x 50;
	
	
	my $cv_final = AnyEvent->condvar;
	push @error, 'apikey is neccessary' unless $args{"apikey"};
	
	
	my $cv_local = AnyEvent->condvar;
	$cv_local->begin (sub { shift->send ( '', \@error ) });
	
	if ($args{apikey}) {
		$cv_local->begin;
		$redis->sismember(
			"apikey.list",
			$args{apikey},
			sub {
				my ($result, $err) = @_;
				push @error, $err if $err;
				push @error, 'apikey check failed' unless $result;
				$cv_local->end;
			}
		);
	}
	
	if ( $args{"layer"} ) {
		$cv_local->begin;
		$redis->sismember(
			$args{apikey} . ".layer.list",
			$args{"layer"},
			sub {
				my ($result, $err) = @_;
				push @error, $err if $err;
				push @error, 'layer check failed' unless $result;
				$cv_local->end;
			}
		);
	}
	
	if ( $args{'name'} ) {
		$cv_local->begin;
		$redis->sismember(
			$args{apikey}.':'.$args{layer}.'.node.list',
			$args{name},
			sub {
				my ($result, $err) = @_;
				push @error, $err if $err;
				push @error, 'name check failed' unless $result;
				$cv_local->end;
			}
		);
	}
	
	$cv_local->end;
	
	
	$cv_local->cb( sub {
		return if $cv_final->ready;
		
		my ($result, $err) = shift->recv;
		
		$result = 'OK' unless @$err;
		
		$cv_final->send( $result, $err );
	});
	
	defined $cb ? $cv_final->cb( sub { $cb->( $_[0]->recv ) } ) : return $cv_final;
}


# list subscibers of 
sub stalkers {
	my $self = shift;
	my $cb = pop if "CODE" eq ref $_[-1];
	my $node_info = shift;
	
	my $redis = $self->{redis};
	my @error;
	
	my $target = $node_info->{apikey} . ':' . $node_info->{layer} . ':' . $node_info->{name};
	
	my $cs = AnyEvent::CallbackStack->new;
	
	$cs->start;
	say "START stalkers" if $DEBUG;
	
	$cs->add( sub {
		$redis->smembers(
			'node:'.$target.'.stalkers',
			sub {
				my ($result, $err) = @_;
				push @error, $err if defined $err;
				$cs->next( $result, \@error );
			},
		);
	});
	
	defined $cb ? $cs->last->cb( sub { $cb->( $_[0]->recv ) } ) : return $cs->last;
	
}


# subscribe someone's update
sub stalk {
	my $self = shift;
	my $cb = pop if "CODE" eq ref $_[-1];
	my $node_info = shift;
	my $target_info = shift;
	
	my $redis = $self->{redis};
	my @error;
	
	my $stalker = $node_info->{apikey} . ':' . $node_info->{layer} . ':' . $node_info->{name};
	my $target = $target_info->{apikey} . ':' . $target_info->{layer} . ':' . $target_info->{name};
	
	say Dumper($stalker, $target) if $DEBUG;
	
	
	my $cs = AnyEvent::CallbackStack->new;
	
	$cs->start;
	say "START stalk" if $DEBUG;
	
	$cs->add( sub {
		$redis->sadd(
			'node:'.$target.'.stalkers', $stalker,
			sub {
				my ($result, $err) = @_;
				push @error, $err if defined $err;
				$result = 'OK' unless @error;
				$cs->next( $result, \@error );
			},
		);
	});
	
	defined $cb ? $cs->last->cb( sub { $cb->( $_[0]->recv ) } ) : return $cs->last;
	
}


# unsubscribe someone's update
sub unstalk {
	my $self = shift;
	my $cb = pop if "CODE" eq ref $_[-1];
	my $node_info = shift;
	my $target_info = shift;
	
	my $redis = $self->{redis};
	my @error;
	
	my $stalker = $node_info->{apikey} . ':' . $node_info->{layer} . ':' . $node_info->{name};
	my $target = $target_info->{apikey} . ':' . $target_info->{layer} . ':' . $target_info->{name};
	
	say Dumper($stalker, $target) if $DEBUG;
	
	
	my $cs = AnyEvent::CallbackStack->new;
	
	$cs->start;
	say "START unstalk" if $DEBUG;
	
	$cs->add( sub {
		$redis->srem(
			'node:'.$target.'.stalkers', $stalker,
			sub {
				my ($result, $err) = @_;
				push @error, $err if defined $err;
				$result = 'OK' unless @error;
				$cs->next( $result, \@error );
			},
		);
	});
	
	defined $cb ? $cs->last->cb( sub { $cb->( $_[0]->recv ) } ) : return $cs->last;
	
}


# remove all data belongs to node/apikey
sub revoke {
	my $self = shift;
	my $cb = pop if "CODE" eq ref $_[-1];
	my %args = @_;
	my $redis = $self->{redis};
	my @error;
	
	say Dumper(%args) if $DEBUG;
	#~ say "la" x 50;
	push @error, 'no input data' unless @_;
	push @error, 'missing data' unless $args{"apikey"} and $args{"layer"};
	
	
	
	my $cv_final = AnyEvent->condvar;
	
	my $cv_local = AnyEvent->condvar;
	$cv_local->begin (sub { shift->send ( '', \@error ) });
	
	if ( $args{"layer"} and $args{'name'} ) {
		
		say "Going to delete Node";
		
		$cv_local->begin;
		$self->_cleanup_info(
			%args,
			sub {
				my ($result, $err) = @_;
				say Dumper($result, $err) if $DEBUG;
				push @error, @$err if @$err;
				push @error, 'node revoke failed' if @$err;
				$cv_local->end;
			}
		);
		# remove from node list under apikey
		$cv_local->begin;
		my $list_prefix = $args{apikey}.':'.$args{layer};
		my $node_key = 'node:' . $args{apikey} . ':' . $args{layer} . ':' . $args{name};
		$redis->srem( "$list_prefix.node.list", $args{'name'}, sub {
			my ($result, $err) = @_;
			push @error, $err if $err;
			$cv_local->end;
		} );
	} elsif ( $args{"layer"} ) {
		
		say "Going to delete Layer";
		
		$cv_local->begin;
		my $list_prefix = $args{apikey}.':'.$args{layer};
		
		my $cs = AnyEvent::CallbackStack->new;
		
		$cs->start;
		
		$cs->add( sub {
			$redis->smembers( "$list_prefix.node.list", sub {
				my ($result, $err) = @_;
				push @error, $err if $err;
				
				$cs->next($result);
			} )
		} );
		
		$cs->add( sub {
			my @node_list = @{ shift->recv };
			
			say "Nodes under the layer: ".Dumper(@node_list) if $DEBUG;
			
			my $cv_for = AnyEvent->condvar;
			$cv_for->begin (sub { shift->send ( '', \@error ) });
			
			for my $name (@node_list) {
				$cv_for->begin;
				$self->revoke(
					%args,
					'name' => $name,
					sub {
						my ($result, $err) = @_;
						push @error, @$err if @$err;
						$cv_for->end;
					}
				);
			}
			
			$cv_for->end;
			
			$cv_for->cb( sub {
				$cs->next;
			});
		});
		
		$cs->add( sub {
			say "Deleting Layer.node.list";
			$redis->del( "$list_prefix.node.list", sub {
				my ($result, $err) = @_;
				push @error, $err if $err;
				$cs->next;
			} );
		});
		
		$cs->add( sub {
			say "Deleting Layer from APIkey.layer.list";
			$redis->srem( $args{"apikey"}.'.layer.list' , $args{"layer"}, sub {
				my ($result, $err) = @_;
				push @error, $err if $err;
				$cs->next;
			} );
		});
		
		# delete
		#~ 1) "list:2aba38af37b17c3d8aa2f22e51b5ca215797d966:theWOWlayer:loc:x.120"
		 #~ 3) "index:2aba38af37b17c3d8aa2f22e51b5ca215797d966:theWOWlayer:loc:z:all"
		 #~ 6) "index:2aba38af37b17c3d8aa2f22e51b5ca215797d966:theWOWlayer:loc:x:all"
		 #~ 9) "index:2aba38af37b17c3d8aa2f22e51b5ca215797d966:theWOWlayer:loc:y:all"
		#~ 10) "list:2aba38af37b17c3d8aa2f22e51b5ca215797d966:theWOWlayer:loc:y.23"
		#~ 11) "list:2aba38af37b17c3d8aa2f22e51b5ca215797d966:theWOWlayer:loc:z.20"
		
		$cs->add( sub {
			$redis->keys( "*$list_prefix*", sub {
				my ($result, $err) = @_;
				push @error, $err if $err;
				
				$cs->next($result);
			} )
		} );
		
		$cs->add( sub {
			my @other_keys = @{ shift->recv };
			say "Deleting other related keys: ".Dumper(@other_keys) if $DEBUG;
			
			my $cv_for = AnyEvent->condvar;
			$cv_for->begin (sub { shift->send ( '', \@error ) });
			
			for my $key (@other_keys) {
				$cv_for->begin;
				$redis->del( $key, sub {
					my ($result, $err) = @_;
					push @error, $err if $err;
					$cv_for->end;
				} );
			}
			
			$cv_for->end;
			
			$cv_for->cb( sub {
				$cs->next;
			});
		});
		
		
		$cs->last->cb( sub {
			$cv_local->end;
		});
	}
	
	$cv_local->end;
	
	
	$cv_local->cb( sub {
		return if $cv_final->ready;
		
		#~ my ($result, $err) = shift->recv;
		
		my $result = 'OK' unless @error;
		
		$cv_final->send( $result, \@error );
	});
	
	defined $cb ? $cv_final->cb( sub { $cb->( $_[0]->recv ) } ) : return $cv_final;
}


# remove node from all lists/indexes and itself
sub _cleanup_info {
	my $self = shift;
	my $cb = pop if "CODE" eq ref $_[-1];
	my %args = @_;
	my $redis = $self->{redis};
	my @error;
	my %node_info;
	my $node_key = 'node:' . $args{apikey} . ':' . $args{layer} . ':' . $args{name};
	
	say "Cleaning Up: ".Dumper(%args) if $DEBUG;
	#~ say "la" x 50;
	push @error, 'no input data' unless @_;
	push @error, 'missing data' unless $args{"apikey"} and $args{"layer"};
	
	
	
	my $cv_final = AnyEvent->condvar;
	my $cs = AnyEvent::CallbackStack->new;
	
	$cs->start(%args);
	say "START _cleanup_info" if $DEBUG;
	
	# get original data
	$cs->add( sub {
		say "START query_node in _cleanup_info" if $DEBUG;
		$self->query_node(
			\%args,
			sub { %node_info = %{$_[0]}; say Dumper %node_info if $DEBUG; $cs->next(@_) },
		);
	});
	
	# remove from list:apikey:layer:loc:x.123 / list:apikey:layer:loc:y.123 / list:apikey:layer:loc:z.123
	$cs->add( sub {
		say Dumper(@_) if $DEBUG;
		my $list_prefix = 'list:' . $args{apikey}.':'.$args{layer} . ':loc';
		my $cv_local = AE::cv;
		
		say 'Removing from lists' if $DEBUG;
		say "$list_prefix: "."x:".$node_info{'loc'}[0]." y:".$node_info{'loc'}[1]." z:".$node_info{'loc'}[2]." $node_key" if $DEBUG;
		
		$cv_local->begin (sub { shift->send ( '', \@error ) });
		$cv_local->begin; $redis->srem( "$list_prefix:x.".$node_info{'loc'}[0], $node_key, sub { my ($result, $err) = @_; push @error, $err if $err; $cv_local->end; } );
		$cv_local->begin; $redis->srem( "$list_prefix:y.".$node_info{'loc'}[1], $node_key, sub { my ($result, $err) = @_; push @error, $err if $err; $cv_local->end; } );
		$cv_local->begin; $redis->srem( "$list_prefix:z.".$node_info{'loc'}[2], $node_key, sub { my ($result, $err) = @_; push @error, $err if $err; $cv_local->end; } );
		$cv_local->end;
		$cv_local->cb( sub {
			my ($result, $err) = shift->recv;
			$cs->next( $result, $err );
		});
	});
	
	# remove node stalker
	$cs->add( sub {
		say "Removing stalkers" if $DEBUG;
		$redis->del(
			$node_key.'.stalkers',
			sub { my ($result, $err) = @_; push @error, $err if $err; $cs->next; }
		);
	});
	
	# remove msgs
	$cs->add( sub {
		say "Removing msgs" if $DEBUG;
		$redis->del(
			$node_key.'.msg.perm',
			sub { my ($result, $err) = @_; push @error, $err if $err; $cs->next; }
		);
	});
	
	$cs->add( sub {
		say "Removing node: $node_key" if $DEBUG;
		$redis->del(
			$node_key,
			sub { my ($result, $err) = @_; push @error, $err if $err; $cs->next; }
		);
	});
	
	$cs->add( sub {
		say "Removing v: $node_key:v" if $DEBUG;
		$redis->del(
			$node_key.":v",
			sub { my ($result, $err) = @_; push @error, $err if $err; $cs->next('', \@error); }
		);
	});
	
	
	$cs->last->cb( sub {
		my ($result, $err) = shift->recv;
		say Dumper($result, $err) if $DEBUG;
		
		$result = 'OK' unless @$err;
		
		$cv_final->send( $result, $err );
	});
	
	defined $cb ? $cv_final->cb( sub { $cb->( $_[0]->recv ) } ) : return $cv_final;
}


1;

